# -*- coding: utf-8 -*-

import torch as th
import torch.nn as nn

from leibniz.nn.layer.cbam import CBAM


class BasicBlock(nn.Module):
    def __init__(self, in_channel, out_channel, step, relu, conv, reduction=16):
        super(BasicBlock, self).__init__()
        self.step = step
        self.relu = relu

        self.conv1 = conv(in_channel, in_channel, kernel_size=3, stride=1, padding=1)
        self.conv2 = conv(in_channel, out_channel, kernel_size=3, stride=1, padding=1)
        self.cbam = CBAM(out_channel, reduction=reduction, conv=conv)

    def forward(self, x):
        y = self.conv1(x)
        y = self.relu(y)
        y = self.conv2(y)
        y = self.cbam(y)
        return y


class Bottleneck(nn.Module):
    def __init__(self, in_channel, out_channel, step, relu, conv, reduction=16):
        super(Bottleneck, self).__init__()
        self.step = step
        self.relu = relu

        self.conv1 = conv(in_channel, in_channel // 4, kernel_size=1, bias=False)
        self.conv2 = conv(in_channel // 4, in_channel // 4, kernel_size=3, bias=False, padding=1)
        self.conv3 = conv(in_channel // 4, out_channel, kernel_size=1, bias=False)
        self.cbam = CBAM(out_channel, reduction=reduction, conv=conv)

    def forward(self, x):
        y = self.conv1(x)
        y = self.relu(y)
        y = self.conv2(y)
        y = self.relu(y)
        y = self.conv3(y)
        y = self.cbam(y)
        return y


class HyperBasic(nn.Module):
    extension = 3
    least_required_dim = 1

    def __init__(self, dim, step, relu, conv, reduction=16):
        super(HyperBasic, self).__init__()
        self.dim = dim // 3
        self.step = step
        self.output = BasicBlock(7 * self.dim, 3 * self.dim, step, relu, conv, reduction=reduction)

    def forward(self, xs):
        x, cs, ss = xs[:, :self.dim], xs[:, self.dim:self.dim*2], xs[:, self.dim*2:]

        y1 = (1 + ss) * x + cs
        y2 = (1 + cs) * x - ss
        y3 = (1 - cs) * x + ss
        y4 = (1 - ss) * x - cs
        ys = th.cat((y1, y2, y3, y4, cs, ss, x), dim=1)

        out = self.output(ys)
        dx, dcs, dss = out[:, :self.dim], out[:, self.dim:self.dim*2], out[:, self.dim*2:]
        x, cs, ss = x + dx, cs + dcs, ss + dss

        return th.cat((x, cs, ss), dim=1)


class HyperBottleneck(nn.Module):
    extension = 3
    least_required_dim = 1

    def __init__(self, dim, step, relu, conv, reduction=16):
        super(HyperBottleneck, self).__init__()
        self.dim = dim // 3
        self.step = step
        self.output = Bottleneck(7 * self.dim, 3 * self.dim, step, relu, conv, reduction=reduction)

    def forward(self, xs):
        x, cs, ss = xs[:, :self.dim], xs[:, self.dim:self.dim*2], xs[:, self.dim*2:]

        y1 = (1 + ss) * x + cs
        y2 = (1 + cs) * x - ss
        y3 = (1 - cs) * x + ss
        y4 = (1 - ss) * x - cs
        ys = th.cat((y1, y2, y3, y4, cs, ss, x), dim=1)

        out = self.output(ys)
        dx, dcs, dss = out[:, :self.dim], out[:, self.dim:self.dim*2], out[:, self.dim*2:]
        x, cs, ss = x + dx, cs + dcs, ss + dss

        return th.cat((x, cs, ss), dim=1)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="leibniz",
    version="0.1.26",
    author="Mingli Yuan",
    author_email="mingli.yuan@gmail.com",
    description="Leibniz is a package providing facilities to express learnable differential equations based on PyTorch",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/caiyunapp/leibniz",
    project_urls={
        'Documentation': 'https://packaging.python.org/tutorials/distributing-packages/',
        'Source': 'https://github.com/caiyunapp/leibniz',
        'Tracker': 'https://github.com/caiyunapp/leibniz/issues',
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
    install_requires=[
        'cached_property',
        'torchpwl',
        'torch',
        'numpy',
    ],
    test_suite='nose.collector',
    tests_require=['nose'],
)

